/*
** Module   : Function.java
** Abstract : contains the data needed to define a Progress function
**
** Copyright (c) 2005-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050413   @20719 First version, simple bean-like approach
**                           including a temporary index to allow the
**                           linking of a function call to its original
**                           definition at parse time. This is needed
**                           because project-unique IDs can't be easily
**                           generated at parse time, but are only added
**                           after the entire tree is complete.
** 002 GES 20050608   @21451 Improvement in annotations for better
**                           function definition conversion.
** 003 GES 20070919   @35135 Added support for the fully qualified name
**                           of the class represented by a FUNC_CLASS.
** 004 ECF 20080704   @39144 Minor memory optimization. Use Boolean
**                           constants instead of instantiating new
**                           Boolean objects.
** 005 GES 20090518   @42407 Import change.
** 006 GES 20110901          Added some class annotations.
** 007 CA  20121212          tempidx for functions is no longer needed, as the
**                           refid is computed during conversion.
*/

package com.goldencode.p2j.uast;

import com.goldencode.ast.*;

/**
 * A simple wrapper class to contain all related data about a function.
 */
class Function
implements ProgressParserTokenTypes,
           TokenDataWrapper
{
   /** The full text name of the function. */
   private String name = null;
   
   /**
    * Parser defined integer constant matching the token type of the
    * function.
    */
   private int tokenType = -1;
   
   /** Defines if this function definition is user-defined or builtin. */
   private boolean builtin = false;
   
   /**
    * Defines if this function can return the unknown value. If set to
    * <code>null</code>, then no information on this topic is known.
    */
   private Boolean returnsUnknown = null;
   
   /** Fully qualified class name of the object instance returned. */
   private String cls = null;
   
   /**
    * Constructs an instance with the required data to be contained.  By
    * default builtin functions have their returns unknown state set to
    * <code>false</code>.  This can be overridden by a subsequent method
    * call to {@link #setReturnsUnknown}.
    *
    * @param    name
    *           The full name of the function.
    * @param    tokenType
    *           The token type (as defined by the parser) of the function.
    * @param    builtin
    *           <code>true</code> if this function is 'built in' to the
    *           Progress runtime.
    * @param    cls
    *           Fully qualified class name of the object instance returned
    *           or <code>null</code> if this function does not return an
    *           object instance.
    */
   public Function(String name, int tokenType, boolean builtin, String cls)
   throws IllegalArgumentException
   {
      if (name == null || tokenType == -1)
         throw new IllegalArgumentException("Invalid name or type.");
      
      this.name      = name;
      this.tokenType = tokenType;
      this.builtin   = builtin;
      this.cls       = cls;
      
      if (builtin)
      {
         returnsUnknown = Boolean.FALSE;
      }
   }
   
   /**
    * Returns the full name of the function.
    *
    * @return   The full name of the function.
    */
   public String getName()
   {
      return name;
   }
   
   /**
    * Sets the full name of the function.
    *
    * @param    name
    *           The full name of the function.
    */
   public void setName(String name)
   {
      this.name = name;
   }
   
   /**
    * Returns the parser-defined token type of the function.
    *
    * @return   The token type of the function.
    */
   public int getTokenType()
   {
      return tokenType;
   }   
   
   /**
    * Sets the parser-defined token type of the function.
    *
    * @param    tokenType
    *           The token type of the function.
    */
   public void setTokenType(int tokenType)
   {
      this.tokenType = tokenType;
   }   
   
   /**
    * This is a no-op, as for function references the associated function
    * definition is computed during conversion.
    *
    * @return   Always returns -1.
    */
   public int getTempIndex()
   {
      return -1;
   }   
   
   /**
    * This is a no-op, as for function references the associated function
    * definition is computed during conversion.
    *
    * @param    tempIdx
    *           The temporary index.
    */
   public void setTempIndex(int tempIdx)
   {
      // no-op
   }
   
   /**
    * Determines if this function definition is user-defined or builtin.
    *
    * @return   <code>true</code> if the function is builtin.
    */
   public boolean isBuiltin()
   {
      return builtin;
   }
   
   /**
    * Sets whether this function definition is user-defined or builtin.
    *
    * @param    builtin
    *           <code>true</code> if the function is builtin.
    */
   public void setBuiltin(boolean builtin)
   {
      this.builtin = builtin;
   }
   
   /**
    * Obtains the fully qualified name of the class of the object instance
    * returned by this function.
    *
    * @return   The fully qualified name of the class of the object instance
    *           returned by this function or <code>null</code> if this
    *           function does not return an object instance.
    */
   public String getClassName()
   {
      return cls;
   }
   
   /**
    * Sets the fully qualified name of the class of the object instance
    * returned by this function.
    *
    * @param    cls
    *           The fully qualified name of the class of the object instance
    *           returned by this function or <code>null</code> if this
    *           function does not return an object instance.
    */
   public void setClassName(String cls)
   {
      this.cls = cls;
   }
   
   /**
    * Determines if this function can return the unknown value.
    *
    * @return   <code>true</code> if the function can return the unknown
    *           value, <code>false</code> if not and <code>null</code> if
    *           this fact is not known.
    */
   public Boolean returnsUnknown()
   {
      return returnsUnknown;
   }
   
   /**
    * Sets whether or not this function can return the unknown value.
    *
    * @param    unknown
    *           <code>true</code> if the function can return the unknown
    *           value, <code>false</code> if not and <code>null</code> if
    *           this fact is not known.
    */
   public void setReturnsUnknown(Boolean unknown)
   {
      this.returnsUnknown = unknown;
   }
   
   /**
    * Stores all non-standard options as annotations to the AST passed as
    * a parameter if this is the original definition, otherwise it stores
    * a subset of values, most important of which is the cross-reference
    * index to allow the original definition to be easily identified and
    * accessed from a reference.  If an option has the default value, it is
    * not saved as an annotation.
    *
    * @param    ast
    *           The AST to annotate.    
    * @param    original
    *           <code>true</code> if this is to annotate the AST of the
    *           original definition, <code>false</code> if the AST is only
    *           a reference.
    * @param    sym
    *           The current symbol resolver.
    */
   public void annotateOptions(Aast ast, boolean original, SymbolResolver sym)
   {
      if (ast == null)
      {
         // nothing to do
         return;
      }
      
      if (tokenType == FUNC_CLASS && cls != null)
      {
         sym.annotateClassRef(cls, ast);
      }
      
      // the following annotations are only stored in the original defining
      // AST rather than in references, unless the reference is to a builtin
      // function which has no defining reference
      if (original || builtin)
      {
         if (original)
         {
            // storing the name and type in annotations makes function
            // definitions easier to generate at conversion time
            ast.putAnnotation("name", name);      
            ast.putAnnotation("type", new Long(tokenType));      
         }      
         
         ast.putAnnotation("builtin", builtin ? Boolean.TRUE : Boolean.FALSE);
         
         if (returnsUnknown != null)
         {
            ast.putAnnotation("returnsunknown", returnsUnknown);
         }      
      }
   }   
}
